@@ -229,7 +229,7 @@ ngx_http_sub_header_filter(ngx_http_request_t *r)
             return ngx_http_next_header_filter(r);
         }
 
-        ctx->matches = ngx_pnalloc(r->pool, sizeof(ngx_array_t));
+        ctx->matches = ngx_palloc(r->pool, sizeof(ngx_array_t));
         if (ctx->matches == NULL) {
             return NGX_ERROR;
         }
@@ -237,7 +237,7 @@ ngx_http_sub_header_filter(ngx_http_request_t *r)
         ctx->matches->elts = matches;
         ctx->matches->nelts = j;
 
-        ctx->tables = ngx_pnalloc(r->pool, sizeof(ngx_http_sub_tables_t));
+        ctx->tables = ngx_palloc(r->pool, sizeof(ngx_http_sub_tables_t));
         if (ctx->tables == NULL) {
             return NGX_ERROR;
         }
@@ -853,12 +853,13 @@ ngx_http_sub_merge_conf(ngx_conf_t *cf, void *parent, void *child)
         conf->pairs = prev->pairs;
         conf->matches = prev->matches;
         conf->tables = prev->tables;
+    }
 
-    } else if (conf->dynamic == 0){
+    if (conf->pairs && conf->dynamic == 0 && conf->tables == NULL) {
         pairs = conf->pairs->elts;
         n = conf->pairs->nelts;
 
-        matches = ngx_pnalloc(cf->pool, sizeof(ngx_http_sub_match_t) * n);
+        matches = ngx_palloc(cf->pool, sizeof(ngx_http_sub_match_t) * n);
         if (matches == NULL) {
             return NGX_CONF_ERROR;
         }
@@ -868,7 +869,7 @@ ngx_http_sub_merge_conf(ngx_conf_t *cf, void *parent, void *child)
             matches[i].value = &pairs[i].value;
         }
 
-        conf->matches = ngx_pnalloc(cf->pool, sizeof(ngx_array_t));
+        conf->matches = ngx_palloc(cf->pool, sizeof(ngx_array_t));
         if (conf->matches == NULL) {
             return NGX_CONF_ERROR;
         }
